@@ -22,6 +22,13 @@
 #include "mmap_cache.h"
 #include "mmap_cache_internals.h"
 
+/* Global time_override */
+MU32 time_override = 0;
+
+void mmc_set_time_override(MU32 set_time) {
+  time_override = set_time;
+}
+
 /* Default values for a new cache */
 char * def_share_file = "/tmp/sharefile";
 MU32    def_init_file = 0;
@@ -40,28 +47,21 @@ MU32    def_start_slots = 89;
  * 
 */
 mmap_cache * mmc_new() {
-  mmap_cache * cache = (mmap_cache *)malloc(sizeof(mmap_cache));
+  mmap_cache * cache = (mmap_cache *)calloc(1, sizeof(mmap_cache));
 
-  cache->mm_var = 0;
-  cache->p_cur = -1;
+  cache->p_cur = NOPAGE;
 
   cache->c_num_pages = def_c_num_pages;
   cache->c_page_size = def_c_page_size;
-  cache->c_size = 0;
 
   cache->start_slots = def_start_slots;
   cache->expire_time = def_expire_time;
 
-  cache->fh = 0;
   cache->share_file = _mmc_get_def_share_filename(cache);
+  cache->permissions = 0640;
   cache->init_file = def_init_file;
   cache->test_file = def_test_file;
 
-  cache->catch_deadlocks = 0;
-  cache->enable_stats = 0;
-
-  cache->last_error = 0;
-
   return cache;
 }
 
@@ -78,6 +78,8 @@ int mmc_set_param(mmap_cache * cache, char * param, char * val) {
     cache->expire_time = atoi(val);
   } else if (!strcmp(param, "share_file")) {
     cache->share_file = val;
+  } else if (!strcmp(param, "permissions")) {
+    cache->permissions = atoi(val);
   } else if (!strcmp(param, "start_slots")) {
     cache->start_slots = atoi(val);
   } else if (!strcmp(param, "catch_deadlocks")) {
@@ -85,8 +87,7 @@ int mmc_set_param(mmap_cache * cache, char * param, char * val) {
   } else if (!strcmp(param, "enable_stats")) {
     cache->enable_stats = atoi(val);
   } else {
-    _mmc_set_error(cache, 0, "Bad set_param parameter: %s", param);
-    return -1;
+    return _mmc_set_error(cache, 0, "Bad set_param parameter: %s", param);
   }
 
   return 0;
@@ -100,8 +101,7 @@ int mmc_get_param(mmap_cache * cache, char * param) {
   } else if (!strcmp(param, "expire_time")) {
     return (int)cache->expire_time;
   } else {
-    _mmc_set_error(cache, 0, "Bad set_param parameter: %s", param);
-    return -1;
+    return _mmc_set_error(cache, 0, "Bad set_param parameter: %s", param);
   }
 }
 
@@ -113,13 +113,13 @@ int mmc_get_param(mmap_cache * cache, char * param) {
  * 
 */
 int mmc_init(mmap_cache * cache) {
-  int i, do_init;
-  MU32 c_num_pages, c_page_size, c_size, start_slots;
+  int i, do_init = cache->init_file;
+  MU32 c_num_pages, c_page_size;
+  MU64 c_size;
 
   /* Need a share file */
   if (!cache->share_file) {
-    _mmc_set_error(cache, 0, "No share file specified");
-    return -1;
+    return _mmc_set_error(cache, 0, "No share file specified");
   }
 
   /* Basic cache params */
@@ -127,12 +127,11 @@ int mmc_init(mmap_cache * cache) {
   ASSERT(c_num_pages >= 1 && c_num_pages <= 1000);
 
   c_page_size = cache->c_page_size;
-  ASSERT(c_page_size >= 1024 && c_page_size <= 16*1024*1024);
+  ASSERT(c_page_size >= 1024 && c_page_size <= 1024*1024*1024);
 
-  start_slots = cache->start_slots;
-  ASSERT(start_slots >= 10 && start_slots <= 500);
+  ASSERT(cache->start_slots >= 10 && cache->start_slots <= 500);
 
-  cache->c_size = c_size = c_num_pages * c_page_size;
+  cache->c_size = c_size = (MU64)c_num_pages * c_page_size;
 
   if ( mmc_open_cache_file(cache, &do_init) == -1) return -1;
 
@@ -141,8 +140,13 @@ int mmc_init(mmap_cache * cache) {
 
   /* Initialise pages if new file */
   if (do_init) {
-    _mmc_init_page(cache, -1);
-    
+    for (i = 0; i < cache->c_num_pages; i++) {
+      MU64 p_offset = (MU64)i * cache->c_page_size;
+      mmc_lock_page(cache, p_offset);
+      _mmc_init_page(cache, i);
+      mmc_unlock_page(cache, p_offset);
+    }
+
     /* Unmap and re-map to stop gtop telling us our memory usage is up */
     if ( mmc_unmap_memory(cache) == -1) return -1;
     if ( mmc_map_memory(cache) == -1) return -1;
@@ -151,23 +155,19 @@ int mmc_init(mmap_cache * cache) {
   /* Test pages in file if asked */
   if (cache->test_file) {
     for (i = 0; i < cache->c_num_pages; i++) {
-      int lock_page = 0, bad_page = 0;
+      int bad_page = 0;
+      MU64 p_offset = (MU64)i * cache->c_page_size;
 
       /* Need to lock page, which tests header structure */
       if (mmc_lock(cache, i)) {
+        /* If that failed, assume bad header, so manually lock */
+        mmc_lock_page(cache, p_offset);
         bad_page = 1;
 
       /* If lock succeeded, test page structure */
       } else {
-        lock_page = 1;
-        if (!_mmc_test_page(cache)) {
+        if (!_mmc_test_page(cache))
           bad_page = 1;
-        }
-      }
-
-      /* If we locked, unlock */
-      if (lock_page) {
-        mmc_unlock(cache);
       }
 
       /* A bad page, initialise it */
@@ -178,6 +178,9 @@ int mmc_init(mmap_cache * cache) {
            are really broken anyway */
         i--;
       }
+
+      mmc_unlock_page(cache, p_offset);
+      cache->p_cur = NOPAGE;
     }
   }
 
@@ -199,10 +202,10 @@ int mmc_close(mmap_cache *cache) {
   ASSERT(cache->mm_var);
 
   /* Shouldn't call if page still locked */
-  ASSERT(cache->p_cur == -1);
+  ASSERT(cache->p_cur == NOPAGE);
 
   /* Unlock any locked page */
-  if (cache->p_cur != -1) {
+  if (cache->p_cur != NOPAGE) {
     mmc_unlock(cache);
   }
 
@@ -215,8 +218,7 @@ int mmc_close(mmap_cache *cache) {
   if (cache->mm_var) {
     res = mmc_unmap_memory(cache);
     if (res == -1) {
-      _mmc_set_error(cache, errno, "Mmap of shared file %s failed", cache->share_file);
-      return -1;
+      return _mmc_set_error(cache, errno, "Mmap of shared file %s failed", cache->share_file);
     }
   }
 
@@ -241,25 +243,29 @@ char * mmc_error(mmap_cache * cache) {
  *
 */
 int mmc_lock(mmap_cache * cache, MU32 p_cur) {
-  MU32 p_offset;
+  MU64 p_offset;
   void * p_ptr;
+  int res = 0;
 
   /* Argument sanity check */
-  if (p_cur > cache->c_num_pages)
-    return -1 + _mmc_set_error(cache, 0, "page %u is larger than number of pages", p_cur);
+  if (p_cur == NOPAGE || p_cur > cache->c_num_pages)
+    return _mmc_set_error(cache, 0, "page %u is NOPAGE or larger than number of pages", p_cur);
 
   /* Check not already locked */
-  if (cache->p_cur != -1)
-    return -1 + _mmc_set_error(cache, 0, "page %u is already locked, can't lock multiple pages", cache->p_cur);
+  if (cache->p_cur != NOPAGE)
+    return _mmc_set_error(cache, 0, "page %u is already locked, can't lock multiple pages", cache->p_cur);
 
   /* Setup page details */
-  p_offset = p_cur * cache->c_page_size;
+  p_offset = (MU64)p_cur * cache->c_page_size;
   p_ptr = PTR_ADD(cache->mm_var, p_offset);
 
-  if (mmc_lock_page(cache, p_offset) == -1) return -1;
+  res = mmc_lock_page(cache, p_offset);
+  if (res) return res;
 
-  if (!(P_Magic(p_ptr) == 0x92f7e3b1))
-    return -1 + _mmc_set_error(cache, 0, "magic page start marker not found. p_cur is %u, offset is %u", p_cur, p_offset);
+  if (!(P_Magic(p_ptr) == 0x92f7e3b1)) {
+    mmc_unlock_page(cache, p_offset);
+    return _mmc_set_error(cache, 0, "magic page start marker not found. p_cur is %u, offset is %llu", p_cur, p_offset);
+  }
 
   /* Copy to cache structure */
   cache->p_num_slots = P_NumSlots(p_ptr);
@@ -272,13 +278,17 @@ int mmc_lock(mmap_cache * cache, MU32 p_cur) {
 
   /* Reality check */
   if (cache->p_num_slots < 89 || cache->p_num_slots > cache->c_page_size)
-    return -1 + _mmc_set_error(cache, 0, "cache num_slots mistmatch");
-  if (cache->p_free_slots < 0 || cache->p_free_slots > cache->p_num_slots)
-    return -1 + _mmc_set_error(cache, 0, "cache free slots mustmatch");
-  if (cache->p_old_slots > cache->p_free_slots)
-    return -1 + _mmc_set_error(cache, 0, "cache old slots mistmatch");
-  if (cache->p_free_data + cache->p_free_bytes != cache->c_page_size)
-    return -1 + _mmc_set_error(cache, 0, "cache free data mistmatch");
+    res = _mmc_set_error(cache, 0, "cache num_slots mistmatch");
+  else if (cache->p_free_slots < 0 || cache->p_free_slots > cache->p_num_slots)
+    res = _mmc_set_error(cache, 0, "cache free slots mustmatch");
+  else if (cache->p_old_slots > cache->p_free_slots)
+    res = _mmc_set_error(cache, 0, "cache old slots mistmatch");
+  else if (cache->p_free_data + cache->p_free_bytes != cache->c_page_size)
+    res = _mmc_set_error(cache, 0, "cache free data mistmatch");
+  if (res) {
+    mmc_unlock_page(cache, p_offset);
+    return res;
+  }
 
   /* Check page header */
   ASSERT(P_Magic(p_ptr) == 0x92f7e3b1);
@@ -289,7 +299,7 @@ int mmc_lock(mmap_cache * cache, MU32 p_cur) {
 
   /* Setup page pointers */
   cache->p_cur = p_cur;
-  cache->p_offset = p_cur * cache->c_page_size;
+  cache->p_offset = p_offset;
   cache->p_base = p_ptr;
   cache->p_base_slots = PTR_ADD(p_ptr, P_HEADERSIZE);
 
@@ -308,7 +318,7 @@ int mmc_lock(mmap_cache * cache, MU32 p_cur) {
 */
 int mmc_unlock(mmap_cache * cache) {
 
-  ASSERT(cache->p_cur != -1);
+  ASSERT(cache->p_cur != NOPAGE);
 
   /* If changed, save page header changes back */
   if (cache->p_changed) {
@@ -322,12 +332,16 @@ int mmc_unlock(mmap_cache * cache) {
     P_FreeBytes(p_ptr) = cache->p_free_bytes;
     P_NReads(p_ptr) = cache->p_n_reads;
     P_NReadHits(p_ptr) = cache->p_n_read_hits;
+
+    cache->p_changed = 0;
   }
 
   /* Test before unlocking */
   ASSERT(_mmc_test_page(cache));
 
-  mmc_unlock_page(cache);
+  mmc_unlock_page(cache, cache->p_offset);
+
+  cache->p_cur = NOPAGE;
 
   return 0;
 }
@@ -342,7 +356,7 @@ int mmc_unlock(mmap_cache * cache) {
 */
 int mmc_is_locked(mmap_cache * cache) {
 
-  return cache->p_cur != -1 ? 1 : 0;
+  return cache->p_cur != NOPAGE ? 1 : 0;
 }
 
 /*
@@ -380,7 +394,7 @@ int mmc_hash(
  *   cache_mmap * cache, MU32 hash_slot,
  *   void *key_ptr, int key_len,
  *   void **val_ptr, int *val_len,
- *   MU32 *flags
+ *   MU32 *expire_on, MU32 *flags
  * )
  *
  * Read key from current page
@@ -390,7 +404,7 @@ int mmc_read(
   mmap_cache *cache, MU32 hash_slot,
   void *key_ptr, int key_len,
   void **val_ptr, int *val_len,
-  MU32 *flags
+  MU32 *expire_on_p, MU32 *flags_p
 ) {
   MU32 * slot_ptr;
 
@@ -413,19 +427,17 @@ int mmc_read(
   } else {
 
     MU32 * base_det = S_Ptr(cache->p_base, *slot_ptr);
-    MU32 now = (MU32)time(0);
+    MU32 now = time_override ? time_override : (MU32)time(0);
 
-    MU32 expire_time = S_ExpireTime(base_det);
+    MU32 expire_on = S_ExpireOn(base_det);
 
     /* Sanity check hash matches */
     ASSERT(S_SlotHash(base_det) == hash_slot);
 
     /* Value expired? */
-    if (expire_time && now > expire_time) {
+    if (expire_on && now >= expire_on) {
 
-      /* Delete slot and return not found */
-      _mmc_delete_slot(cache, slot_ptr);
-      ASSERT(*slot_ptr == 1);
+      /* Return not found, but leave slot. Might need writeback */
 
       return -1;
     }
@@ -434,13 +446,16 @@ int mmc_read(
     S_LastAccess(base_det) = now;
 
     /* Copy values to pointers */
-    *flags = S_Flags(base_det);
+    *flags_p = S_Flags(base_det);
+    *expire_on_p = expire_on;
     *val_len = S_ValLen(base_det);
     *val_ptr = S_ValPtr(base_det);
 
     /* Increase read hit count */
-    if (cache->enable_stats)
+    if (cache->enable_stats) {
+      cache->p_changed = 1;
       cache->p_n_read_hits++;
+    }
 
     return 0;
   }
@@ -451,7 +466,7 @@ int mmc_read(
  *   cache_mmap * cache, MU32 hash_slot,
  *   void *key_ptr, int key_len,
  *   void *val_ptr, int val_len,
- *   MU32 expire_seconds, MU32 flags
+ *   MU32 expire_on, MU32 flags
  * )
  *
  * Write key to current page
@@ -461,7 +476,7 @@ int mmc_write(
   mmap_cache *cache, MU32 hash_slot,
   void *key_ptr, int key_len,
   void *val_ptr, int val_len,
-  MU32 expire_seconds, MU32 flags
+  MU32 expire_on, MU32 flags
 ) {
   int did_store = 0;
   MU32 kvlen = KV_SlotLen(key_len, val_len);
@@ -475,7 +490,7 @@ int mmc_write(
 
   ROUNDLEN(kvlen);
 
-  ASSERT(cache->p_cur != -1);
+  ASSERT(cache->p_cur != NOPAGE);
 
   /* If found, delete slot for new value */
   if (*slot_ptr > 1) {
@@ -488,16 +503,15 @@ int mmc_write(
   /* If there's space, store the key/value in the data section */
   if (cache->p_free_bytes >= kvlen) {
     MU32 * base_det = PTR_ADD(cache->p_base, cache->p_free_data);
-    MU32 now = (MU32)time(0);
-    MU32 expire_time = 0;
+    MU32 now = time_override ? time_override : (MU32)time(0);
 
     /* Calculate expiry time */
-    if (expire_seconds == (MU32)-1) expire_seconds = cache->expire_time;
-    expire_time = expire_seconds ? now + expire_seconds : 0;
+    if (expire_on == (MU32)-1)
+      expire_on = cache->expire_time ? now + cache->expire_time : 0;
 
     /* Store info into slot */
     S_LastAccess(base_det) = now;
-    S_ExpireTime(base_det) = expire_time;
+    S_ExpireOn(base_det) = expire_on;
     S_SlotHash(base_det) = hash_slot;
     S_Flags(base_det) = flags;
     S_KeyLen(base_det) = (MU32)key_len;
@@ -598,7 +612,7 @@ int mmc_calc_expunge(
 ) {
   double slots_pct;
 
-  ASSERT(cache->p_cur != -1);
+  ASSERT(cache->p_cur != NOPAGE);
 
   /* If len >= 0, and space available for len bytes, nothing is expunged */
   if (len >= 0) {
@@ -621,20 +635,20 @@ int mmc_calc_expunge(
     MU32 * slot_end = slot_ptr + num_slots;
 
     /* Store pointers to used slots */
-    MU32 ** copy_base_det = (MU32 **)malloc(sizeof(MU32 *) * used_slots);
+    MU32 ** copy_base_det = (MU32 **)calloc(used_slots, sizeof(MU32 *));
     MU32 ** copy_base_det_end = copy_base_det + used_slots;
     MU32 ** copy_base_det_out = copy_base_det;
     MU32 ** copy_base_det_in = copy_base_det + used_slots;
 
     MU32 page_data_size = cache->c_page_size - num_slots * 4 - P_HEADERSIZE;
     MU32 in_slots, data_thresh, used_data = 0;
-    MU32 now = (MU32)time(0);
+    MU32 now = time_override ? time_override : (MU32)time(0);
 
     /* Loop for each existing slot, and store in a list */
     for (; slot_ptr != slot_end; slot_ptr++) {
       MU32 data_offset = *slot_ptr;
       MU32 * base_det = S_Ptr(cache->p_base, data_offset);
-      MU32 expire_time, flags, kvlen;
+      MU32 expire_on, kvlen;
 
       /* Ignore if if free slot */
       if (data_offset <= 1) {
@@ -648,9 +662,8 @@ int mmc_calc_expunge(
       }
 
       /* Definitely out if expired, and not dirty */
-      expire_time = S_ExpireTime(base_det);
-      flags = S_Flags(base_det);
-      if (expire_time && now >= expire_time) {
+      expire_on = S_ExpireOn(base_det);
+      if (expire_on && now >= expire_on) {
         *copy_base_det_out++ = base_det;
         continue;
       }
@@ -733,13 +746,17 @@ int mmc_do_expunge(
 
   /* Build new slots data and KV data */
   MU32 slot_data_size = new_num_slots * 4;
-  MU32 * new_slot_data = (MU32 *)malloc(slot_data_size);
+  MU32 * new_slot_data = (MU32 *)calloc(1, slot_data_size);
 
   MU32 page_data_size = cache->c_page_size - new_num_slots * 4 - P_HEADERSIZE;
 
-  void * new_kv_data = malloc(page_data_size);
+  void * new_kv_data = calloc(1, page_data_size);
   MU32 new_offset = 0;
 
+  /* Sanity check underlying fd is still the same file */
+  if (!mmc_check_fh(cache))
+    return 0;
+
   /* Start all new slots empty */
   memset(new_slot_data, 0, slot_data_size);
 
@@ -806,7 +823,7 @@ int mmc_do_expunge(
 
   ASSERT(_mmc_test_page(cache));
 
-  return 0;
+  return 1;
 }
 
 /*
@@ -843,11 +860,9 @@ void mmc_reset_page_details(mmap_cache * cache) {
  *
 */
 mmap_cache_it * mmc_iterate_new(mmap_cache * cache) {
-  mmap_cache_it * it = (mmap_cache_it *)malloc(sizeof(mmap_cache_it));
+  mmap_cache_it * it = (mmap_cache_it *)calloc(1, sizeof(mmap_cache_it));
   it->cache = cache;
-  it->p_cur = -1;
-  it->slot_ptr = 0;
-  it->slot_ptr_end = 0;
+  it->p_cur = NOPAGE;
 
   return it;
 }
@@ -865,23 +880,28 @@ MU32 * mmc_iterate_next(mmap_cache_it * it) {
   mmap_cache * cache = it->cache;
   MU32 * slot_ptr = it->slot_ptr;
   MU32 * base_det;
+  MU32 expire_on;
+  MU32 now = time_override ? time_override : (MU32)time(0);
 
-  /* If empty slot, keep moving till we find a used one */
-  while (slot_ptr == it->slot_ptr_end || *slot_ptr <= 1) {
+  /* Go until we find a slot or exit */
+  while (1) {
 
     /* End of page ... */
     if (slot_ptr == it->slot_ptr_end) {
 
-      /* Unlock current page if any */
-      if (it->p_cur != -1) {
-        mmc_unlock(it->cache);
-      }
+      if (it->p_cur == NOPAGE) {
+        it->p_cur = 0;
 
-      /* Move to the next page, return 0 if no more pages */
-      if (++it->p_cur == cache->c_num_pages) {
-        it->p_cur = -1;
-        it->slot_ptr = 0;
-        return 0;
+      /* Unlock current page if any */
+      } else {
+        mmc_unlock(it->cache);
+
+        /* Move to the next page, return 0 if no more pages */
+        if (++it->p_cur == cache->c_num_pages) {
+          it->p_cur = NOPAGE;
+          it->slot_ptr = 0;
+          return 0;
+        }
       }
 
       /* Lock the new page number */
@@ -895,12 +915,24 @@ MU32 * mmc_iterate_next(mmap_cache_it * it) {
       continue;
     }
 
-    /* Move to next slot */
-    slot_ptr++;
-  }
+    /* Slot not used */
+    if (*slot_ptr <= 1) {
+      slot_ptr++;
+      continue;
+    }
 
-  /* Get pointer to details for this entry */
-  base_det = S_Ptr(cache->p_base, *slot_ptr);
+    /* Get pointer to details for this entry */
+    base_det = S_Ptr(cache->p_base, *slot_ptr);
+
+    /* Slot expired */
+    expire_on = S_ExpireOn(base_det);
+    if (expire_on && now >= expire_on) {
+      slot_ptr++;
+      continue;
+    }
+
+    break;
+  }
 
   /* Move to the next slot for next iteration */
   it->slot_ptr = ++slot_ptr;
@@ -917,7 +949,7 @@ MU32 * mmc_iterate_next(mmap_cache_it * it) {
 */
 void mmc_iterate_close(mmap_cache_it * it) {
   /* Unlock page if locked */
-  if (it->p_cur != -1) {
+  if (it->p_cur != NOPAGE) {
     mmc_unlock(it->cache);
   }
 
@@ -931,7 +963,7 @@ void mmc_iterate_close(mmap_cache_it * it) {
  *   MU32 * base_det,
  *   void ** key_ptr, int * key_len,
  *   void ** val_ptr, int * val_len,
- *   MU32 * last_access, MU32 * expire_time, MU32 * flags
+ *   MU32 * last_access, MU32 * expire_on, MU32 * flags
  * )
  *
  * Given a base_det pointer to entries details
@@ -945,7 +977,7 @@ void mmc_get_details(
   MU32 * base_det,
   void ** key_ptr, int * key_len,
   void ** val_ptr, int * val_len,
-  MU32 * last_access, MU32 * expire_time, MU32 * flags
+  MU32 * last_access, MU32 * expire_on, MU32 * flags
 ) {
   cache = cache;
 
@@ -956,7 +988,7 @@ void mmc_get_details(
   *val_len = S_ValLen(base_det);
 
   *last_access = S_LastAccess(base_det);
-  *expire_time = S_ExpireTime(base_det);
+  *expire_on = S_ExpireOn(base_det);
   *flags = S_Flags(base_det);
 }
 
@@ -973,7 +1005,7 @@ void _mmc_delete_slot(
   mmap_cache * cache, MU32 * slot_ptr
 ) {
   ASSERT(*slot_ptr > 1);
-  ASSERT(cache->p_cur != -1);
+  ASSERT(cache->p_cur != NOPAGE);
 
   /* Set offset to 1 */
   *slot_ptr = 1;
@@ -1011,7 +1043,7 @@ MU32 * _mmc_find_slot(
   slots_left = cache->p_num_slots;
   slots_end = cache->p_base_slots + slots_left;
 
-  ASSERT(cache->p_cur != -1);
+  ASSERT(cache->p_cur != NOPAGE);
 
   /* Loop with integer probing till we find or don't */
   while (slots_left--) {
@@ -1025,9 +1057,9 @@ MU32 * _mmc_find_slot(
     /* data_offset == 1 means deleted slot, we can reuse if writing */
     if (data_offset == 0) {
       /* Return pointer to last checked slot */
-      return slot_ptr;
+      break;
     }
-    if (data_offset == 1 && mode == 1 && 0 == slot_ptr) {
+    if (data_offset == 1 && mode == 1 && 0 == first_deleted) {
       /* Save pointer to first usable slot; if we don't find the key later,
          we'll fall back to returning this.
       */
@@ -1055,46 +1087,41 @@ MU32 * _mmc_find_slot(
     if (++slot_ptr == slots_end) { slot_ptr = cache->p_base_slots; }
     ASSERT(slot_ptr >= cache->p_base_slots && slot_ptr < slots_end);
   }
+  /* No slot found */
+  if (++slots_left == 0) slot_ptr = 0;
 
   if (1 == mode && 0 != first_deleted)
     return first_deleted;
   else
-    return 0;
+    return slot_ptr;
 }
 
 /*
  * void _mmc_init_page(mmap_cache * cache, int page)
  *
- * Initialise the given page as empty
- *
- * If page == -1, init all pages
+ * Initialise the given page as empty. It's expected
+ *  that you've already locked the page before doing
+ *  this
  *
 */
 void _mmc_init_page(mmap_cache * cache, MU32 p_cur) {
-  MU32 start_page = p_cur, end_page = p_cur+1;
-  if (p_cur == -1) {
-    start_page = 0;
-    end_page = cache->c_num_pages;
-  }
+  /* Setup page details */
+  MU64 p_offset = (MU64)p_cur * cache->c_page_size;
+  void * p_ptr = PTR_ADD(cache->mm_var, p_offset);
 
-  for (p_cur = start_page; p_cur < end_page; p_cur++) {
-    /* Setup page details */
-    MU32 p_offset = p_cur * cache->c_page_size;
-    void * p_ptr = PTR_ADD(cache->mm_var, p_offset);
+  /* Initialise to all 0's */
+  memset(p_ptr, 0, cache->c_page_size);
 
-    /* Initialise to all 0's */
-    memset(p_ptr, 0, cache->c_page_size);
+  /* Setup header */
+  P_Magic(p_ptr) = 0x92f7e3b1;
+  P_NumSlots(p_ptr) = cache->start_slots;
+  P_FreeSlots(p_ptr) = cache->start_slots;
+  P_OldSlots(p_ptr) = 0;
+  P_FreeData(p_ptr) = P_HEADERSIZE + cache->start_slots * 4;
+  P_FreeBytes(p_ptr) = cache->c_page_size - P_FreeData(p_ptr);
+  P_NReads(p_ptr) = 0;
+  P_NReadHits(p_ptr) = 0;
 
-    /* Setup header */
-    P_Magic(p_ptr) = 0x92f7e3b1;
-    P_NumSlots(p_ptr) = cache->start_slots;
-    P_FreeSlots(p_ptr) = cache->start_slots;
-    P_OldSlots(p_ptr) = 0;
-    P_FreeData(p_ptr) = P_HEADERSIZE + cache->start_slots * 4;
-    P_FreeBytes(p_ptr) = cache->c_page_size - P_FreeData(p_ptr);
-    P_NReads(p_ptr) = 0;
-    P_NReadHits(p_ptr) = 0;
-  }
 }
 
 /*
@@ -1108,8 +1135,8 @@ int  _mmc_test_page(mmap_cache * cache) {
   MU32 count_free = 0, count_old = 0, max_data_offset = 0;
   MU32 data_size = cache->c_page_size;
 
-  ASSERT(cache->p_cur != -1);
-  if (!(cache->p_cur != -1)) return 0;
+  ASSERT(cache->p_cur != NOPAGE);
+  if (cache->p_cur == NOPAGE) return 0;
 
   for (; slot_ptr < cache->p_base_slots + cache->p_num_slots; slot_ptr++) {
     MU32 data_offset = *slot_ptr;
@@ -1133,16 +1160,16 @@ int  _mmc_test_page(mmap_cache * cache) {
       MU32 * base_det = S_Ptr(cache->p_base, data_offset);
 
       MU32 last_access = S_LastAccess(base_det);
-      MU32 expire_time = S_ExpireTime(base_det);
+      MU32 expire_on = S_ExpireOn(base_det);
       MU32 key_len = S_KeyLen(base_det);
       MU32 val_len = S_ValLen(base_det);
       MU32 kvlen = S_SlotLen(base_det);
       ROUNDLEN(kvlen);
 
-      ASSERT(last_access > 1000000000 && last_access < 1500000000);
-      if (!(last_access > 1000000000 && last_access < 1500000000)) return 0;
-      ASSERT(expire_time == 0 || (expire_time > 1000000000 && expire_time < 1500000000));
-      if (!(expire_time == 0 || (expire_time > 1000000000 && expire_time < 1500000000))) return 0;
+      ASSERT(last_access > 1000000000);
+      if (!(last_access > 1000000000)) return 0;
+      ASSERT(expire_on == 0 || (expire_on > 1000000000));
+      if (!(expire_on == 0 || (expire_on > 1000000000))) return 0;
 
       ASSERT(key_len >= 0 && key_len < data_size);
       if (!(key_len >= 0 && key_len < data_size)) return 0;
@@ -1195,7 +1222,7 @@ int  _mmc_test_page(mmap_cache * cache) {
 int _mmc_dump_page(mmap_cache * cache) {
   MU32 slot;
 
-  ASSERT(cache->p_cur != -1);
+  ASSERT(cache->p_cur != NOPAGE);
 
   printf("PageNum: %d\n", cache->p_cur);
   printf("\n");
@@ -1221,7 +1248,7 @@ int _mmc_dump_page(mmap_cache * cache) {
       char key[256], val[256];
 
       printf("LA=%d, ET=%d, HS=%d, FL=%d\n",
-        S_LastAccess(base_det), S_ExpireTime(base_det),
+        S_LastAccess(base_det), S_ExpireOn(base_det),
         S_SlotHash(base_det), S_Flags(base_det));
 
       /* Get data */
